from typing import Any, Union

import jax
import jax.numpy as jnp
import jaxlib
import numpy as np


# JAX arrays
Array = Union[jax.core.Tracer, jaxlib.xla_extension.DeviceArray]
# JAX arrays + numpy arrays
# jnp.ndarray counts as a numpy array: `assert isinstance(np.array(1.), jnp.ndarray)`
MoreArrays = Union[Array, np.ndarray, jnp.ndarray]

PyTree = Any

TreeDef = jaxlib.xla_extension.PyTreeDef



import jax

from .custom_types import PyTree


def _apply_update(p, u):
    if u is None:
        return p
    else:
        return p + u


def apply_updates(model: PyTree, updates: PyTree) -> PyTree:
    return jax.tree_map(_apply_update, model, updates)



import math
from typing import Optional

import jax.random as jrandom

from ..custom_types import Array
from ..module import Module


class Linear(Module):
    weight: Array
    bias: Optional[Array]

    def __init__(self, in_features, out_features, use_bias=True, *, key):
        super().__init__()
        wkey, bkey = jrandom.split(key, 2)
        lim = 1 / math.sqrt(in_features)
        self.weight = jrandom.uniform(
            wkey, (out_features, in_features), minval=-lim, maxval=lim
        )
        if use_bias:
            self.bias = jrandom.uniform(bkey, (out_features,), minval=-lim, maxval=lim)
        else:
            self.bias = None

    def __call__(self, x, *, key=None):
        x = self.weight @ x
        if self.bias is not None:
            x = x + self.bias
        return x


class Identity(Module):
    def __init__(self, *args, **kwargs):
        # Ignores args and kwargs
        super().__init__()

    def __call__(self, x, *, key=None):
        return x



import functools as ft
from dataclasses import dataclass
from typing import Any

import jax

from .filters import validate_filters


@ft.lru_cache(maxsize=4096)
def _jitf_cache(f, args_treedef, **jitkwargs):
    @ft.partial(jax.jit, **jitkwargs)
    def f_wrapped(*args):
        args = jax.tree_unflatten(args_treedef, args)
        return f(*args)

    return f_wrapped


@dataclass(frozen=True)
class _UnPyTreeAble:
    value: Any

    def __bool__(self):
        return False


_marker_sentinel = object()


def jitf(
    fun,
    *,
    filter_fn=None,
    filter_tree=None,
    static_argnums=None,
    static_argnames=None,
    donate_argnums=(),
    **jitkwargs
):
    """
    A jax.jit that automatically sets whether arguments are static or not, according to either `filter_fn` or
    `filter_tree`. The `static_argnums` argument can still be used to additionally specify any extra static arguments.

    The above applies recursively inside PyTrees, e.g. some parts of the PyTree can be static and some can be traced.
    """
    if isinstance(static_argnums, int):
        static_argnums = (static_argnums,)
    if static_argnames is not None:
        raise NotImplementedError(
            "jitf does not yet support `static_argnames`. use static_argnums instead."
        )
    if donate_argnums != ():
        raise NotImplementedError("jitf does not ye support `donate_argnums`.")
    validate_filters("jitf", filter_fn, filter_tree)

    if static_argnums is None:
        len_static_argnums = 0
    else:
        len_static_argnums = len(static_argnums)

    @ft.wraps(fun)
    def f_wrapper(*args, **kwargs):
        if len(kwargs):
            raise NotImplementedError(
                "jitf does not yet support keyword arguments. Use positional arguments instead."
            )

        if filter_tree is not None:
            if len(args) - len_static_argnums == 1:
                new_filter_tree = (filter_tree,)
            else:
                new_filter_tree = tuple(filter_tree)

        # Mark the arguments that have been explicitly declared static via `static_argnums`
        if static_argnums is not None:
            args = list(args)
            for index in static_argnums:
                args[index] = _UnPyTreeAble(args[index])
            if filter_tree is not None:
                new_filter_tree = list(new_filter_tree)
                for index in static_argnums:
                    new_filter_tree.insert(index, _UnPyTreeAble(None))

        # Flatten everything else
        args_flat, args_treedef = jax.tree_flatten(args)
        if filter_tree is not None:
            filter_flat, flat_treedef = jax.tree_flatten(new_filter_tree)
            if flat_treedef != args_treedef:
                raise ValueError(
                    "The tree stucture for the filters and the arguments must be the same."
                )

        # Figure out static argnums with respect to this new flattened structure.
        new_static_argnums = []
        if filter_tree is None:
            # implies filter_fn is not None
            for i, arg in enumerate(args_flat):
                if isinstance(arg, _UnPyTreeAble) or not filter_fn(arg):
                    new_static_argnums.append(i)
        else:
            for i, (arg, filter) in enumerate(zip(args_flat, filter_flat)):
                if not filter:
                    new_static_argnums.append(i)
        new_static_argnums = tuple(new_static_argnums)
        if static_argnums is not None:
            args_flat = [
                arg.value if isinstance(arg, _UnPyTreeAble) else arg
                for arg in args_flat
            ]

        f_jitted = _jitf_cache(
            fun, args_treedef, static_argnums=new_static_argnums, **jitkwargs
        )
        return f_jitted(*args_flat)

    return f_wrapper



from typing import Any, Callable, List, Optional, Tuple
from typing_extensions import get_args

import jax
import jax.numpy as jnp

from .custom_types import Array, MoreArrays, PyTree, TreeDef


_array_types = get_args(Array)
_morearray_types = get_args(MoreArrays)
_arraylike_types = _morearray_types + (int, float, complex, bool)


# TODO: not sure if this is the best way to do this? In light of:
# https://github.com/google/jax/commit/258ae44303b1539eff6253263694ec768b8803f0#diff-de759f969102e9d64b54a299d11d5f0e75cfe3052dc17ffbcd2d43b250719fb0
def is_array(element: Any) -> bool:
    return isinstance(element, _array_types)


# Does _not_ do a try/except on jnp.asarray(element) because that's very slow.
def is_array_like(element: Any) -> bool:
    return isinstance(element, _arraylike_types)


def is_inexact_array(element: Any) -> bool:
    return is_array(element) and jnp.issubdtype(element.dtype, jnp.inexact)


def is_inexact_array_like(element: Any) -> bool:
    return (
        isinstance(element, _morearray_types)
        and jnp.issubdtype(element.dtype, jnp.inexact)
    ) or isinstance(element, (float, complex))


def split(
    pytree: PyTree,
    filter_fn: Optional[Callable[[Any], bool]] = None,
    filter_tree: Optional[PyTree] = None,
) -> Tuple[List[Any], List[Any], List[bool], TreeDef]:

    validate_filters("split", filter_fn, filter_tree)
    flat, treedef = jax.tree_flatten(pytree)
    flat_true = []
    flat_false = []

    if filter_fn is None:
        which, treedef_filter = jax.tree_flatten(filter_tree)
        if treedef != treedef_filter:
            raise ValueError(
                "filter_tree must have the same tree structure as the PyTree being split."
            )
        for f, w in zip(flat, which):
            if w:
                flat_true.append(f)
            else:
                flat_false.append(f)
    else:
        which = []
        for f in flat:
            if filter_fn(f):
                flat_true.append(f)
                which.append(True)
            else:
                flat_false.append(f)
                which.append(False)

    return flat_true, flat_false, which, treedef


def merge(
    flat_true: List[Any], flat_false: List[Any], which: List[bool], treedef: TreeDef
):
    flat = []
    flat_true = iter(flat_true)
    flat_false = iter(flat_false)
    for element in which:
        if element:
            flat.append(next(flat_true))
        else:
            flat.append(next(flat_false))
    return jax.tree_unflatten(treedef, flat)


def validate_filters(fn_name, filter_fn, filter_tree):
    if (filter_fn is None and filter_tree is None) or (
        filter_fn is not None and filter_tree is not None
    ):
        raise ValueError(
            f"Precisely one of `filter_fn` and `filter_tree` should be passed to {fn_name}"
        )



import jax.numpy as jnp
import jax.random as jrandom

from ..module import Module


class Dropout(Module):
    p: float = 0.5
    deterministic: bool = False

    def __call__(self, x, *, key=None, deterministic=None):
        if deterministic is None:
            deterministic = self.deterministic
        if deterministic:
            return x
        elif key is None:
            raise RuntimeError(
                "Dropout requires a key when running in non-deterministic mode."
            )
        else:
            q = 1 - self.p
            mask = jrandom.bernoulli(key, q, x.shape)
            return jnp.where(mask, x / q, 0)



from typing import List

import jax.nn as jnn
import jax.random as jrandom

from ..module import Module
from .linear import Linear


class MLP(Module):
    layers: List[Linear]
    activation: callable
    final_activation: callable

    def __init__(
        self,
        in_size,
        out_size,
        width_size,
        depth,
        activation=jnn.relu,
        final_activation=lambda x: x,
        *,
        key,
        **kwargs
    ):
        super().__init__(**kwargs)
        keys = jrandom.split(key, depth + 1)
        layers = []
        if depth == 0:
            layers.append(Linear(in_size, out_size, key=keys[0]))
        else:
            layers.append(Linear(in_size, width_size, key=keys[0]))
            for i in range(depth - 1):
                layers.append(Linear(width_size, width_size, key=keys[i + 1]))
            layers.append(Linear(width_size, out_size, key=keys[-1]))
        self.layers = layers
        self.activation = activation
        self.final_activation = final_activation

    def __call__(self, x, *, key=None):
        for layer in self.layers[:-1]:
            x = layer(x)
            x = self.activation(x)
        x = self.layers[-1](x)
        x = self.final_activation(x)
        return x


class Sequential(Module):
    layers: List[Module]

    def __call__(self, x, *, key=None):
        if key is None:
            keys = [None] * len(self.layers)
        else:
            keys = jrandom.split(key, len(self.layers))
        for layer, key in zip(self.layers, keys):
            x = layer(x, key=key)
        return x



import functools as ft

import jax

from .filters import merge, split, validate_filters


def value_and_grad_f(fun, *, filter_fn=None, filter_tree=None, argnums=0, **gradkwargs):
    if isinstance(argnums, int):
        unwrap = True
        argnums = (argnums,)
        if filter_tree is not None:
            filter_tree = (filter_tree,)
    else:
        unwrap = False

    validate_filters("value_and_grad_f", filter_fn, filter_tree)

    @ft.partial(jax.value_and_grad, argnums=argnums, **gradkwargs)
    def f_value_and_grad(*args, **kwargs):
        *args, notes = args
        args = list(args)
        for i, (arg_nograd, which, treedef) in notes.items():
            arg_grad = args[i]
            arg = merge(arg_grad, arg_nograd, which, treedef)
            args[i] = arg
        return fun(*args, **kwargs)

    def f_value_and_grad_wrapper(*args, **kwargs):
        args = list(args)
        notes = {}
        for j, i in enumerate(argnums):
            arg = args[i]
            if filter_fn is None:
                # implies filter_tree is not None
                arg_grad, arg_nograd, which, treedef = split(
                    arg, filter_tree=filter_tree[j]
                )
            else:
                arg_grad, arg_nograd, which, treedef = split(arg, filter_fn=filter_fn)
            args[i] = arg_grad
            notes[i] = (arg_nograd, which, treedef)
        value, grad = f_value_and_grad(*args, notes, **kwargs)
        grad = list(grad)
        for j, i in enumerate(argnums):
            g = grad[j]
            arg_nograd, which, treedef = notes[i]
            none_grad = [None for _ in arg_nograd]
            grad[j] = merge(g, none_grad, which, treedef)
        if unwrap:
            (grad,) = grad
        return value, grad

    return f_value_and_grad_wrapper


def gradf(fun, *, has_aux=False, **gradkwargs):
    f_value_and_grad = value_and_grad_f(fun, has_aux=has_aux, **gradkwargs)

    def f_grad(*args, **kwargs):
        value, grad = f_value_and_grad(*args, **kwargs)
        if has_aux:
            value, aux = value
            return aux, grad
        else:
            return grad

    return f_grad



import pathlib
import re

import setuptools


_here = pathlib.Path(__file__).resolve().parent

name = "equinox"

# for simplicity we actually store the version in the __version__ attribute in the
# source
with open(_here / name / "__init__.py") as f:
    meta_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.M)
    if meta_match:
        version = meta_match.group(1)
    else:
        raise RuntimeError("Unable to find __version__ string.")

author = "Patrick Kidger"

author_email = "contact@kidger.site"

description = "PyTorch-like neural networks in JAX"

with open(_here / "README.md", "r") as f:
    readme = f.read()

url = "https://github.com/patrick-kidger/" + name

license = "Apache-2.0"

classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: Financial and Insurance Industry",
    "Intended Audience :: Information Technology",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Natural Language :: English",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Scientific/Engineering :: Mathematics",
]

python_requires = "~=3.7"

install_requires = ["jax>=0.2.18", "typing_extensions"]

setuptools.setup(
    name=name,
    version=version,
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    description=description,
    long_description=readme,
    long_description_content_type="text/markdown",
    url=url,
    license=license,
    classifiers=classifiers,
    zip_safe=False,
    python_requires=python_requires,
    install_requires=install_requires,
    packages=setuptools.find_packages(exclude=["examples", "tests"]),
)



from typing import Any, Callable, Optional, Tuple, Union
from typing_extensions import get_args

import jax

from .custom_types import MoreArrays, PyTree


_sentinel = object()


def tree_at(
    where: Callable[[PyTree], Union[Any, Tuple[Any, ...]]],
    pytree: PyTree,
    replace: Optional[Union[Any, Tuple[Any, ...]]] = _sentinel,
    replace_fn: Optional[Callable[[Any], Any]] = _sentinel,
) -> PyTree:

    if (replace is _sentinel and replace_fn is _sentinel) or (
        replace is not _sentinel and replace_fn is not _sentinel
    ):
        raise ValueError(
            "Precisely one of `replace` and `replace_fn` must be specified."
        )
    elif replace is _sentinel:
        replace_passed = False
        replacer = lambda j, i: replace_fn(flat[i])
    else:
        replace_passed = True
        replacer = lambda j, i: replace[j]

    # TODO: is there a neater way of accomplishing this?
    flat, treedef = jax.tree_flatten(pytree)
    flat_indices = list(range(len(flat)))
    index_pytree = jax.tree_unflatten(treedef, flat_indices)
    index = where(index_pytree)
    # where can return either a single entry, or a tuple.
    if isinstance(index, int):
        index = (index,)
        replace = (replace,)

    if replace_passed and len(index) != len(replace):
        raise ValueError("`where` must return a tuple of the same length as `replace`.")
    for j, i in enumerate(index):
        flat[i] = replacer(j, i)

    return jax.tree_unflatten(treedef, flat)


def tree_equal(*pytrees: PyTree) -> bool:
    flat, treedef = jax.tree_flatten(pytrees[0])
    array_types = get_args(MoreArrays)
    for pytree in pytrees[1:]:
        flat_, treedef_ = jax.tree_flatten(pytree)
        if treedef_ != treedef:
            return False
        for elem, elem_ in zip(flat, flat_):
            if isinstance(elem, array_types):
                if isinstance(elem_, array_types):
                    if (
                        (type(elem) != type(elem_))
                        or (elem.shape != elem_.shape)
                        or (elem.dtype != elem_.dtype)
                        or (elem != elem_).any()
                    ):
                        return False
                else:
                    return False
            else:
                if isinstance(elem_, array_types):
                    return False
                else:
                    if elem != elem_:
                        return False
    return True



import abc
from dataclasses import dataclass, fields

import jax

from .tree import tree_equal


# dataclasses.astuple operates recursively, which destroys information about
# nested tree_dataclasses. In contrast this is just a shallow tuplification.
def _dataclass_astuple(cls):
    return tuple(getattr(cls, field.name) for field in fields(cls) if field.init)


def _allow_setattr(fields):
    def __setattr__(self, name, value):
        if name in fields:
            object.__setattr__(self, name, value)
        else:
            raise AttributeError(f"Cannot set attribute {name}")

    return __setattr__


# Inherits from abc.ABCMeta as a convenience for a common use-case.
# It's not a feature we use ourselve.
class _ModuleMeta(abc.ABCMeta):
    def __new__(mcs, name, bases, dict_):
        try:
            user_provided_init = dict_["__init__"]
        except KeyError:
            reinstate_init = False
        else:
            reinstate_init = True
            del dict_["__init__"]

        cls = super().__new__(mcs, name, bases, dict_)
        cls = dataclass(eq=False, frozen=True)(cls)

        assert "__dataclass_init__" not in cls.__dict__
        assert "__dataclass_setattr__" not in cls.__dict__
        cls.__dataclass_init__ = cls.__init__
        cls.__dataclass_setattr__ = cls.__setattr__
        if not reinstate_init:
            # Override the default dataclass init if our parent has an init
            for kls in cls.__mro__[1:-1]:
                if (
                    isinstance(kls, _ModuleMeta)
                    and kls.__init__ is not kls.__dataclass_init__
                ):
                    reinstate_init = True
                    user_provided_init = kls.__init__
                    break
        if reinstate_init:
            cls.__init__ = user_provided_init

        def flatten(self):
            return _dataclass_astuple(self), None

        def unflatten(_, fields):
            self = cls.__new__(cls, *fields)
            cls.__dataclass_init__(self, *fields)
            return self

        jax.tree_util.register_pytree_node(cls, flatten, unflatten)
        return cls

    def __call__(cls, *args, **kwargs):
        self = cls.__new__(cls, *args, **kwargs)
        # Defreeze it during __init__. TODO: this isn't thread/recursion-safe.
        field_names = {field.name for field in fields(cls)}
        cls.__setattr__ = _allow_setattr(field_names)
        cls.__init__(self, *args, **kwargs)
        cls.__setattr__ = cls.__dataclass_setattr__
        missing_names = {name for name in field_names if name not in dir(self)}
        if len(missing_names):
            raise ValueError(
                f"The following fields were not initialised during __init__: {missing_names}"
            )
        return self


class Module(metaclass=_ModuleMeta):
    def __eq__(self, other):
        return tree_equal(self, other)



import math
from typing import Optional

import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom

from ..custom_types import Array
from ..module import Module


class GRUCell(Module):
    weight_ih: Array
    weight_hh: Array
    bias: Optional[Array]
    bias_n: Optional[Array]

    def __init__(self, input_size, hidden_size, bias=True, *, key, **kwargs):
        super().__init__(**kwargs)

        ihkey, hhkey, bkey, bkey2 = jrandom.split(key, 4)
        lim = math.sqrt(1 / hidden_size)

        self.weight_ih = jrandom.uniform(
            ihkey, (3 * hidden_size, input_size), minval=-lim, maxval=lim
        )
        self.weight_hh = jrandom.uniform(
            hhkey, (3 * hidden_size, hidden_size), minval=-lim, maxval=lim
        )
        if bias:
            self.bias = jrandom.uniform(
                bkey, (3 * hidden_size,), minval=-lim, maxval=lim
            )
            self.bias_n = jrandom.uniform(
                bkey2, (hidden_size,), minval=-lim, maxval=lim
            )
        else:
            self.bias = None
            self.bias_n = None

    def __call__(self, input, hidden, *, key=None):
        if self.bias is None:
            bias = 0
            bias_n = 0
        else:
            bias = self.bias
            bias_n = self.bias_n
        igates = jnp.split(self.weight_ih @ input + bias, 3)
        hgates = jnp.split(self.weight_hh @ hidden, 3)
        reset = jnn.sigmoid(igates[0] + hgates[0])
        inp = jnn.sigmoid(igates[1] + hgates[1])
        new = jnn.tanh(igates[2] + reset * (hgates[2] + bias_n))
        return new + inp * (hidden - new)


class LSTMCell(Module):
    weight_ih: Array
    weight_hh: Array
    bias: Optional[Array]

    def __init__(self, input_size, hidden_size, bias=True, *, key, **kwargs):
        super().__init__(**kwargs)

        ihkey, hhkey, bkey = jrandom.split(key, 3)
        lim = math.sqrt(1 / hidden_size)

        self.weight_ih = jrandom.uniform(
            ihkey, (4 * hidden_size, input_size), minval=-lim, maxval=lim
        )
        self.weight_hh = jrandom.uniform(
            hhkey, (4 * hidden_size, hidden_size), minval=-lim, maxval=lim
        )
        if bias is None:
            self.bias = jrandom.uniform(
                bkey, (4 * hidden_size,), minval=-lim, maxval=lim
            )
        else:
            self.bias = None

    def __call__(self, input, hidden, *, key=None):
        h, c = hidden
        lin = self.weight_ih @ input + self.weight_hh @ h
        if self.bias is not None:
            lin = lin + self.bias
        i, f, g, o = jnp.split(lin, 4)
        i = jnn.sigmoid(i)
        f = jnn.sigmoid(f)
        g = jnn.tanh(g)
        o = jnn.sigmoid(o)
        c = f * c + i * g
        h = o * jnn.tanh(c)
        return (h, c)



from . import nn
from .filters import (
    is_array,
    is_array_like,
    is_inexact_array,
    is_inexact_array_like,
    merge,
    split,
)
from .gradf import gradf, value_and_grad_f
from .jitf import jitf
from .module import Module
from .tree import tree_at, tree_equal
from .update import apply_updates


__version__ = "0.0.3"
